/* lldrv.c 
 * kernel driver for LittleLights / BlinkenLedsPro (based on drive for BLINKENmini)
 *  - character device kernel module for input from and output to LittleLights / BlinkenLedsPro using the parallel port
 *  - version 1.1 date 2005-01-03
 * Copyright (C) 2004-2005 ST <st@littlelights.de> & 1stein <1stein@schuermans.info>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/param.h>
#include <linux/interrupt.h>

//number of elements in an array
#define count( array ) (sizeof( (array) ) / sizeof( (array)[0] ))

//version information for insmod
char version[] = UTS_RELEASE;

//startup parameters
int io = 0;
MODULE_PARM( io, "i" );
int irq = 0;
MODULE_PARM( irq, "i" );

//include lldrv-header-file
#include "lldrv.h"

//global constants concerning the module
#define LlDrvDevName "lldrv" //the name of this device (max. 31 chars allowed)
#define LlDrvDevId ((void *)(('l'<<24)+('l'<<16)+('d'<<8)+'r')) //the id of this device
#define LlDrvDevVerMaj 1 //version of this device
#define LlDrvDevVerMin 0
#define LlDrvDevVerRev 0

//global types
#define LlDrvInternalPictureSizeX 18 //dimensions of an internal BM-picture
#define LlDrvInternalPictureSizeY 8
typedef struct tLlDrvInternalPicture //an internal BM-picture
{
  unsigned char Pixels[LlDrvInternalPictureSizeY][LlDrvInternalPictureSizeX];
} stLlDrvInternalPicture;

//global variables
int LlDrvIsOpen = 0; //true if the device is opened by a process
unsigned short LlDrvIoAddrs[] = { 0x3BC, 0x378, 0x278 }; //valid io-addresses
unsigned short LlDrvIoAddr; //io-address
unsigned char LlDrvIrqNos[] = { 7, 5 }; //valid irq-numbers
unsigned char LlDrvIrqNo; //irq-number

//brightness-level to duty-cycle conversion
#define LlDrvOutputSteps 15
unsigned char LlDrvOutputStepPixelVal[LlDrvOutputSteps] =
{ 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70, 0x80, 0x90, 0xA0, 0xB0, 0xC0, 0xD0, 0xE0, 0xF0 };
//the next step of the output-process to take when an interrupt occurs
int LlDrvOutputStep = 0;

//the current picture that is being displayed
struct tLlDrvInternalPicture LlDrvCurrentPicture;

//the next picture to display
//this picture exists two times
//one can be altered by the ioctl-syscall
//the other one can be read by the interrupt (which can happen at any time)
struct tLlDrvInternalPicture LlDrvNextPictures[2];
//the index of the picture that currently belongs to the interrupt
int LlDrvNextPictureInterruptIndex;

//the standard-picture: ll-logo.bmm
struct tLlDrvInternalPicture LlDrvStdPicture =
{
Pixels:
  {
    { 0x11, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff, 0x00, 0x88 },
    { 0x22, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x77 },
    { 0x33, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x66 },
    { 0x44, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x55 },
    { 0x55, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x44 },
    { 0x66, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x33 },
    { 0x77, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x22 },
    { 0x88, 0x00, 0xff, 0xee, 0xdd, 0xcc, 0xbb, 0xaa, 0x99, 0x88, 0x77, 0x66, 0x55, 0x44, 0x33, 0x22, 0x11, 0x11 }
   }
};

//global variables for reading from the device
//format is 4 bytes per event:
//  0xFF <action> <controller-no> <button-no>
//  action: 0x00: button released, 0x01: button pressed
//  example: 0xFF 0x01 0x00 0x02 ---> button 2 of controller 0 was pressed
#define LlDrvReadBufferSize 64
unsigned char LlDrvReadBuffer[LlDrvReadBufferSize][4]; //ring-buffer with actions
int LlDrvReadBufferStart; //start-index for writing into buffer
int LlDrvReadBufferEnd; //end-index for writing into buffer
unsigned char LlDrvReadChrBuffer[4]; //buffer for left characters from the last read
int LlDrvReadChrBufferCnt; //number of characters in buffer for left characters from the last read

//output a frame
//turns on all pixels of *pPicture that have a value >= PixelValue
//if pPicture is NULL, all pixels are turned off
void LlDrvFrameOut( struct tLlDrvInternalPicture * pPicture, unsigned char PixelValue )
{
  //local variables
  int X, Y;
  unsigned char Val, Strobe;

  //loop through all 9 bit groups
  for( X = 1; X >= 0; X-- )
  for( Y = 7; Y >= 0; Y-- )
  {
    //calculate value and strobe for this 9 pixels
    Val = 0x00;
    Strobe = 0x00;
    if( pPicture != NULL )
    {
      if( pPicture->Pixels[Y][ 0 + X] >= PixelValue )
        Val |= 0x01;
      if( pPicture->Pixels[Y][ 2 + X] >= PixelValue )
        Val |= 0x02;
      if( pPicture->Pixels[Y][ 4 + X] >= PixelValue )
        Val |= 0x04;
      if( pPicture->Pixels[Y][ 6 + X] >= PixelValue )
        Val |= 0x08;
      if( pPicture->Pixels[Y][ 8 + X] >= PixelValue )
        Val |= 0x10;
      if( pPicture->Pixels[Y][10 + X] >= PixelValue )
        Val |= 0x20;
      if( pPicture->Pixels[Y][12 + X] >= PixelValue )
        Val |= 0x40;
      if( pPicture->Pixels[Y][14 + X] >= PixelValue )
        Val |= 0x80;
      if( pPicture->Pixels[Y][16 + X] >= PixelValue )
        Strobe |= 0x01;
    }

    //output the next 9 bits of data
    outb( Val, LlDrvIoAddr );
    outb( 0x13 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

    //output a clock-pulse (on pin -nAUTOLF)
    outb( 0x11 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=0, nINIT=0, -nSLCTIN=0
    outb( 0x13 & ~Strobe, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=!Strobe, -nAUTOLF=1, nINIT=0, -nSLCTIN=0
  } //for( Y ... for( X ...

  //output a load-pulse (on pin nINIT)
  outb( 0x17, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=1, -nSLCTIN=0
  outb( 0x13, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0
}

//the interrupt procedure
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
irqreturn_t LlDrvInt( int IrqNo, void * DevId, struct pt_regs * PtRegs )
#else
void LlDrvInt( int IrqNo, void * DevId, struct pt_regs * PtRegs )
#endif
{
  //check IRQ-number and device-id
  if( IrqNo != LlDrvIrqNo || DevId != LlDrvDevId )
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
    return IRQ_NONE;
#else
    return;
#endif

  //start of a new output-cycle
  if( LlDrvOutputStep == 0 )
    //copy next picture (that belongs to the interrupt) to current picture
    memcpy( &LlDrvCurrentPicture, &LlDrvNextPictures[LlDrvNextPictureInterruptIndex], sizeof( struct tLlDrvInternalPicture ) );

  //output the current frame of the current picture
  LlDrvFrameOut( &LlDrvCurrentPicture, LlDrvOutputStepPixelVal[LlDrvOutputStep] );

  //next output step
  LlDrvOutputStep++;
  if( LlDrvOutputStep >= LlDrvOutputSteps )
    LlDrvOutputStep = 0;

#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  return IRQ_HANDLED;
#endif
}

//a process is opening the device
int LlDrvFileOpen( struct inode * Inode, struct file * File )
{
  //if output device is already open
  if( LlDrvIsOpen )
    //open failed
    return -EBUSY;

  //device is now open
  LlDrvIsOpen = 1;

  //empty next picture that belongs not to the interrupt
  memset( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], 0, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;

  //open succeeded
  return 0;  
}

//a process is closing the device
int LlDrvFileClose( struct inode * Inode, struct file * File )
{
  //if device is not open
  if( !LlDrvIsOpen )
    //close failed
    return -EBADF;

  //copy standard picture into next picture that belongs not to the interrupt
  memcpy( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], &LlDrvStdPicture, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
 
  //device is now closed
  LlDrvIsOpen = 0;

  //close succeeded
  return 0;  
}

//a process is writing to the device
ssize_t LlDrvFileWrite( struct file * File, const char * p_buffer, size_t count, loff_t * offset_ptr )
{
  int cnt, channels, maxval, x, y, xx, yy;
  int fill_l, fill_r, fill_t, fill_b, skip_l, skip_t, startpos;
  unsigned long magic;
  unsigned short width, height;
  unsigned char val;

  //if output device is not open
  if( !LlDrvIsOpen )
    //write failed
    return -EBADF;

  //test buffer
  if( verify_area( VERIFY_READ, p_buffer, count ) != 0 )
    //buffer not okay
    return -EFAULT;

  //default: only one byte written
  // - idea is to skip only the first byte if the data could not be understood
  if( count >= 1 )
    cnt = 1;
  else
    cnt = 0;

  //get magic
  magic = 0;
  if( count >= 4 )
  {
    //get magic
    get_user( magic, (unsigned long *)p_buffer );
    magic = ntohl( magic );
  }
  //get parameters
  switch( magic )
  {
    //DEADBEEF and FEEDBEEF frame
    case 0xDEADBEEF:
    case 0xFEEDBEEF:
      //check length of packet
      if( count < 12 )
      {
        magic = 0;
        width = 0;
        height = 0;
        channels = 1;
        maxval = 1;
        startpos = 4;
        break;
      }
      //get width and height
      get_user( width, (unsigned short *)(p_buffer+8) );
      width = ntohs( width );
      get_user( height, (unsigned short *)(p_buffer+10) );
      height = ntohs( height );
      //set rest of parameters
      channels = 1;
      maxval = magic == 0xFEEDBEEF ? 0xFF : 1;
      startpos = 12;
      break;
    //MCU frame
    case 0x23542666:
      //check length of packet
      if( count < 12 )
      {
        magic = 0;
        width = 0;
        height = 0;
        channels = 1;
        maxval = 1;
        startpos = 4;
        break;
      }
      //get height and widtht
      get_user( height, (unsigned short *)(p_buffer+4) );
      height = ntohs( height );
      get_user( width, (unsigned short *)(p_buffer+6) );
      width = ntohs( width );
      //get channels and maxval
      get_user( channels, (unsigned short *)(p_buffer+8) );
      channels = ntohs( channels );
      get_user( maxval, (unsigned short *)(p_buffer+10) );
      maxval = ntohs( maxval );
      //set rest of parameters
      startpos = 12;
      break;
    default:
      magic = 0;
      width = 0;
      height = 0;
      channels = 1;
      maxval = 1;
      startpos = 4;
      break;
  }
  //some valid packet
  if( magic != 0 )
  {
    //check length of packet
    if( count >= startpos + (int)width * (int)height * (int)channels )
    {
      //calculate parameters
      fill_l = width < LlDrvInternalPictureSizeX ? (LlDrvInternalPictureSizeX - width) / 2 : 0;
      fill_r = width < LlDrvInternalPictureSizeX ? LlDrvInternalPictureSizeX - width - fill_l : 0;
      fill_t = height < LlDrvInternalPictureSizeY ? (LlDrvInternalPictureSizeY - height) / 2 : 0;
      fill_b = height < LlDrvInternalPictureSizeY ? LlDrvInternalPictureSizeY - height - fill_t : 0;
      skip_l = width > LlDrvInternalPictureSizeX ? (width - LlDrvInternalPictureSizeX) / 2 : 0;
      skip_t = height > LlDrvInternalPictureSizeY ? (height - LlDrvInternalPictureSizeY) / 2 : 0;
      //write frame into buffer for next picture
      for( y = 0; y < LlDrvInternalPictureSizeY; y++ )
      for( x = 0; x < LlDrvInternalPictureSizeX; x++ )
      {
        //default: off
        LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex].Pixels[y][x] = 0;
        //real pixel
        if( x >= fill_l && x < LlDrvInternalPictureSizeX - fill_r && y >= fill_t && y < LlDrvInternalPictureSizeY - fill_b )
        {
          xx = x - fill_l + skip_l;
          yy = y - fill_t + skip_t;
          get_user( val, p_buffer + startpos + (yy * width + xx) * channels );
          val = val > maxval ? 0xFF : (unsigned char)((unsigned long)(0xFF * val) / maxval);
          LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex].Pixels[y][x] = val;
        }
      }
      //remember number of bytes written to this device
      cnt = startpos + (int)width * (int)height * (int)channels;
      //make the new next picture the one that belongs to the interrupt
      //(because the assignment is an atomic instruction nothing can go wrong)
      //(the interrupt only reads this variable)
      //(this line can't be executed during execution of the interrupt)
      LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;
    }
  }

  //return number of bytes written to this device
  return cnt;
}

//a process is reading from the output device
//reading from this device is always nonblocking - so when read returning 0 does not mean EOF
ssize_t LlDrvFileRead( struct file * File, char * BufferPtr, size_t BufferLen, loff_t * OffsetPtr )
{
  //if output device is not open
  if( !LlDrvIsOpen )
    //read failed
    return -EBADF;

  //test buffer
  if( verify_area( VERIFY_WRITE, BufferPtr, BufferLen ) != 0 )
    //buffer not okay
    return -EFAULT;

  //nothing written
  return 0;
}

//structure for supported operations
struct file_operations LlDrvFileOperations =
{
  .owner = THIS_MODULE,
  .read = LlDrvFileRead,
  .write = LlDrvFileWrite,
  .open = LlDrvFileOpen,
  .release = LlDrvFileClose,
  .llseek = NULL,
  .readdir = NULL,
  .poll = NULL,
  .ioctl = NULL,
  .mmap = NULL,
  .flush = NULL,
#if LINUX_VERSION_CODE >= 0x20600 //kernel-version 2.6
  .aio_read = NULL,
  .aio_write = NULL,
  .fsync = NULL,
  .aio_fsync = NULL,
  .fasync = NULL,
  .lock = NULL,
  .readv = NULL,
  .writev = NULL,
  .sendpage = NULL,
  .get_unmapped_area = NULL,
  .sendfile = NULL,
#endif
};

//module is being loaded
int init_module( )
{
  //local variables
  int I;
  struct resource * ResPtr;

  //short message
  printk( "lldrv: LittleLights driver for the parallel port (version %d.%d.%d)\n", LlDrvDevVerMaj, LlDrvDevVerMin, LlDrvDevVerRev );
  printk( "lldrv: done in 2004 by ST <st@littlelights.de> & 1stein <1stein@schuermans.info>\n" );
  printk( "lldrv: distributed under GNU Public License - no warranty\n" );
  printk( "lldrv: trying to load (io=0x%X irq=%d)...\n", io, irq );

  //check startup parameters - io-address
  for( I = 0; I < count( LlDrvIoAddrs ); I++ )
    if( LlDrvIoAddrs[I] == (unsigned short)io )
      break;
  //invalid io-address
  if( I >= count( LlDrvIoAddrs ) )
  {
    //error message
    printk( "lldrv: io=0x%X is not valid\n", io );
    printk( "lldrv: valid settings for io are:" );
    for( I = 0; I < count( LlDrvIoAddrs ); I++ )
      printk( " 0x%X", LlDrvIoAddrs[I] );
    printk( "\n" );
    //init failed
    return -1;
  }
  //remember used io-address
  LlDrvIoAddr = LlDrvIoAddrs[I];

  //check startup parameters - irq-number
  for( I = 0; I < count( LlDrvIrqNos ); I++ )
    if( LlDrvIrqNos[I] == (unsigned char)irq )
      break;
  //invalid irq-number
  if( I >= count( LlDrvIrqNos ) )
  {
    //error message
    printk( "lldrv: irq=0x%X is not valid\n", irq );
    printk( "lldrv: valid settings for irq are:" );
    for( I = 0; I < count( LlDrvIrqNos ); I++ )
      printk( " %d", LlDrvIrqNos[I] );
    printk( "\n" );
    //init failed
    return -1;
  }
  //remember used irq-number
  LlDrvIrqNo = LlDrvIrqNos[I];

  //output and input device is not opend by a process yet
  LlDrvIsOpen = 0;

  //set the next picture to the standard picture
  //(nothing can go wrong, because interrupt is not active)
  memcpy( &LlDrvNextPictures[0], &LlDrvStdPicture, sizeof( struct tLlDrvInternalPicture ) ); //set picture 0
  LlDrvNextPictureInterruptIndex = 0; //picture 0 belongs to interrupt
  //start outputting the next picture
  //(nothing can go wrong, because interrupt is not active)
  LlDrvOutputStep = 0;

  //register device
  I = register_chrdev( LlDrvDevNoMajor, LlDrvDevName, &LlDrvFileOperations );
  //an error occured
  if( I < 0 )
  {
    //error-message
    printk( "lldrv: register_chrdev failed with %d\n", I );
    //init failed
    return -1;
  }

  //register io-ports
  ResPtr = request_region( LlDrvIoAddr, 3, LlDrvDevName );
  //not free
  if( ResPtr == NULL )
  {
    //error-message
    printk( "lldrv: request_region failed: io-ports not free\n" );
    //unregister device
    unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );
    //init failed
    return -1;
  }

  //setup hardware
  outb( 0x00, LlDrvIoAddr ); //set data-bits to 0
  outb( 0x13, LlDrvIoAddr + 2 ); //bidirectional: off, irq: on, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

  //register hardware-interrupt (not shareable)
  I = request_irq( LlDrvIrqNo, LlDrvInt, 0, LlDrvDevName, LlDrvDevId );
  //an error occured
  if( I != 0 )
  {
    //error-message
    printk( "lldrv: request_irq failed with %d\n", I );
    //unregister io-ports
    release_region( LlDrvIoAddr, 3 );
    //unregister device
    unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );
    //init failed
    return -1;
  }

  //init succeeded
  printk( "lldrv: loaded (io=0x%X irq=%d)...\n", LlDrvIoAddr, LlDrvIrqNo );
  return 0;
}

//module is being unloaded
void cleanup_module( )
{
  //empty next picture that belongs not to the interrupt
  memset( &LlDrvNextPictures[1 - LlDrvNextPictureInterruptIndex], 0, sizeof( struct tLlDrvInternalPicture ) );
  //make the new next picture the one that belongs to the interrupt
  //(because the assignment is an atomic instruction nothing can go wrong)
  //(the interrupt only reads this variable)
  //(this line can't be executed during execution of the interrupt)
  LlDrvNextPictureInterruptIndex = 1 - LlDrvNextPictureInterruptIndex;

  //unregister hardware-interrupt
  free_irq( LlDrvIrqNo, LlDrvDevId );

  //output empty frame
  LlDrvFrameOut( NULL, 0 );

  //reset hardware
  outb( 0x00, LlDrvIoAddr ); //set data-bits to 0
  outb( 0x03, LlDrvIoAddr + 2 ); //bidirectional: off, irq: off, -nSTROBE=1, -nAUTOLF=1, nINIT=0, -nSLCTIN=0

  //unregister io-ports
  release_region( LlDrvIoAddr, 3 );

  //unregister device
  unregister_chrdev( LlDrvDevNoMajor, LlDrvDevName );

  //done
  printk( "lldrv: unloaded...\n" );
}

